"""
Provides an interface to run a number of different Simple Climate Models.

In future, this will likely be replaced with _OpenSCM as that project aims to provide a consistent interface to a wider variety of
models.

It is recommended that the scm objects are used within a context manager (see below). This helps clean up any temporary resources
used. This is particularly important for the parallel versions of MAGICC (``MAGICC6_Parallel`` and ``MAGICC7_Parallel``) which
may create a large number of instances and processes.

    .. code-block:: python

        from scmcallib.scm import MAGICC7_SCM

        with MAGICC7_SCM() as m:
            # A new magicc instance is only created on the first call to ``run``
            res = m.run({}, variables=['Surface Temperature'])
            # Reuses the same instance
            res = m.run({}, variables=['Surface Temperature'])

        # the magicc instance is automatically removed once out of the scope of the context manager.


    .. _openscm: https://github.com/openclimatedata/openscm
"""

# TODO: This module could likely be replaced with openscm once it is stable
from .base import BaseSCM  # noqa: F401
from .ar5ir import AR5IR
from .magicc import MAGICC7_SCM, MAGICC6_SCM, MAGICC6_Parallel, MAGICC7_Parallel

_models = [AR5IR, MAGICC6_SCM, MAGICC7_SCM, MAGICC6_Parallel, MAGICC7_Parallel]


def get_scm(scm_name):
    """
    Get a SCM class

    Parameters
    ----------
    scm_name: str
        SCM name (``name`` attribute on the class)

    Returns
    -------
    An SCM class which can then be initialised

    """
    for m in _models:
        if m.name == scm_name.lower():
            return m

    raise ValueError(
        "No scm named {}. Options include {}".format(
            scm_name, [m.name for m in _models]
        )
    )



"""
Transforms are arbitary functions which perform some transformation on an input data set and return the result. In the optimisation
cycle, these transformations are applied to the raw model output before any metrics are calculated.
This allows for penalising for values outside a range or performing subsampling.

At a high-level the optimisation cycle is as follows:

#. Optimiser produces parameters
#. Model run
#. Output transformed
#. Metrics calculated
#. Repeat until convergence
"""

import ast


def penalise(lower=None, upper=None, value=1e9):
    """
    Applies a penalty to values outside a range

    Parameters
    ----------
    lower: int
        Values lower than this are penalised
    upper: int
        Values more than `upper` are penalised
    value: float
        The value to set all the items outside the valid range

    """

    def _wrapper(df):
        mask = False

        if lower is not None:
            mask = mask | (df < lower)
        if upper is not None:
            mask = mask | (df > upper)
        df[mask] = value
        return df

    return _wrapper


"""
Each transform is a function which can take arguments and keywords and returns a wrapped function which performs the
actual transformation
"""
transforms = {"penalise": penalise}


def _dump_ast(c):
    if isinstance(c, ast.Num):
        return c.n
    elif isinstance(c, ast.Str):
        return c.s
    elif isinstance(c, ast.Tuple) or isinstance(c, ast.List):
        return tuple([_dump_ast(i) for i in c.elts])
    elif isinstance(c, ast.UnaryOp) and isinstance(c.op, ast.USub):
        return -_dump_ast(c.operand)
    else:
        raise SyntaxError()


def _parse_transform(name):
    if "(" not in name:
        if ")" in name:
            raise SyntaxError()
        return name, [], {}

    # Use ast to parse the function call
    # Much simpler than using regex
    call = ast.parse(name)

    # Only parse simple functions (no modules) of form func(arg1, arg2, .., argn, kw1=val1, .., kwn=val2)
    # Only handles numbers and strings or list/tuples of them as args and kwargs as no other state will be available
    if len(call.body) != 1 and isinstance(call.body[0].value, ast.Call):
        raise ValueError("Cannot parse function call from transform: {}".format(name))

    call = call.body[0].value
    func_name = call.func.id

    args = [_dump_ast(i) for i in call.args]
    kwargs = {i.arg: _dump_ast(i.value) for i in call.keywords}

    return func_name, args, kwargs


def get_transform(name, *args, **kwargs):
    """
    Lookup a transform function

    The transform must be registered in ``transforms``

    Parameters
    ----------
    name: str
        Transform name.

        Can be in the form of a function call, such as ``func_name(arg1, arg2, kw1=val1)`` in which case the args and kwargs are
        parsed from call.
    args
        Transform args. Only used if a function name is passed to name, instead of a function call.
    kwargs
        Transform kwargs. Only used if a function name is passed to name, instead of a function call.
    Returns
    -------
    Function which takes a ``pd.DataFrame`` and returns a modified DataFrame
    """
    name, f_args, f_kwargs = _parse_transform(name)

    if len(f_args):
        if len(args):
            raise ValueError("Can't provide args in both function and as arguments")
        args = f_args

    if len(f_kwargs):
        if len(kwargs):
            raise ValueError("Can't provide kwargs in both function and as arguments")
        kwargs = f_kwargs

    if name not in transforms:
        raise ValueError("Unknown transform: {}".format(name))

    return transforms[name](*args, **kwargs)



from sklearn.metrics import (
    explained_variance_score,
    mean_absolute_error,
    mean_squared_error,
    mean_squared_log_error,
    median_absolute_error,
    r2_score,
)


# To align with PyMC3 and BayesOpt, we want to always maximize a metric.
# Hence if the metric is an error (like MSE), ensure it is negative.
# As scipy can only minimise, we will need to make the error terms positive again (
# i.e. take a double negative) before passing them to scipy. Hence our backend is a
# bit convoluted, but the user experience is clean.


def neg(f):
    def _wrapper(y, y_pred, sample_weight=None):
        return -f(y, y_pred, sample_weight=sample_weight)

    return _wrapper


def _med_abs_err(y_true, y_pred, sample_weight=None):
    return median_absolute_error(y_true, y_pred)


LARGE_NEG_ERR = -1e9

metrics = {
    "r_squared": r2_score,
    "explained_variance": explained_variance_score,
    "neg_mean_sq_err": neg(mean_squared_error),
    "neg_mean_abs_err": neg(mean_absolute_error),
    "neg_mean_sq_log_err": neg(mean_squared_log_error),
    "neg_median_abs_err": neg(_med_abs_err),
}


def get_metric(metric):
    """Get the function for the metric used to determine goodness of fit

    The returned function returns a single float goodness of fit value from two input
    vectors.

    Notes
    -----
    - More information about the available metrics is available from
        https://scikit-learn.org/stable/modules/model_evaluation.html#regression-metrics

    Parameters
    ----------
    metric :
        Case insensitive string containing the selected metric. Options include

        - r_squared: R^2 value
        - explained_variance: Measure to which the modelled data accounts for the variation
        - neg_mean_sq_err: Mean square error (MSE)
        - neg_mean_abs_err: Mean absolute error (MAE)
        - neg_mean_sq_log_err: Mean square log error (MSLE)
        - neg_median_abs_err: Median absolute error (MedAE)

        The signature for the metric functions are ``(y, y_pred, sample_weight=None) -> float`` where y is the

        - y_true : array-like of shape = (n_samples)
            Ground truth (correct) target values.
        - y_pred : array-like of shape = (n_samples)
            Estimated target values.
        - sample_weight : array-like of shape = (n_samples), optional
            Sample weights.

    Returns
    -------
    function
        A function taking the sample time series and the expected time series as arguments. The functions all return a single float.

    """
    try:
        return metrics[metric]
    except KeyError:
        raise ValueError("No metric named: {}".format(metric))



import shutil
from concurrent.futures import ProcessPoolExecutor
from logging import getLogger
from multiprocessing import current_process, Manager
from os.path import exists, join
from subprocess import CalledProcessError
from tempfile import mkdtemp

import pandas as pd
from scmdata import ScmDataFrame, df_append
from pymagicc import MAGICC6, MAGICC7
from pymagicc.io import MAGICCData

from scmcallib.utils import parallel_process
from .base import BaseSCM

logger = getLogger(__name__)

# Set the EMIS and Tuning configuration as pymagicc2.0 expects
config_defaults = {
    "file_emisscen_2": "NONE",
    "file_emisscen_3": "NONE",
    "file_emisscen_4": "NONE",
    "file_emisscen_5": "NONE",
    "file_emisscen_6": "NONE",
    "file_emisscen_7": "NONE",
    "file_emisscen_8": "NONE",
    "file_tuningmodel_1": "PYMAGICC",
    "file_tuningmodel_2": "USER",
    "file_tuningmodel_3": "USER",
    "file_tuningmodel_4": "USER",
    "file_tuningmodel_5": "USER",
    "file_tuningmodel_6": "USER",
    "file_tuningmodel_7": "USER",
    "file_tuningmodel_8": "USER",
    "file_tuningmodel_9": "USER",
    "file_tuningmodel_10": "USER",
}


def _generate_magicc_root(root_dir):
    return mkdtemp(prefix="pymagicc-", dir=root_dir)


_instances = {}


def get_magicc_inst(magicc_version=7, root_dir=None, init_callback=None):
    """Gets a MAGICC object which is ready to run

    This caches the magicc instance used to minimise overhead from copying files. Each process gets a unique copy of MAGICC
    to ensure that each chain has exclusive access to the magicc instance.

    Returns
    -------
    pymagicc.MAGICC7
        MAGICC7 object with a valid configuration
    """

    key = (magicc_version, current_process().name)
    try:
        return _instances[key]
    except KeyError:
        kwargs = {}
        if root_dir:
            kwargs["root_dir"] = _generate_magicc_root(root_dir)

        if magicc_version == 6:
            magicc = MAGICC6(**kwargs)
        else:
            magicc = MAGICC7(strict=False, **kwargs)
        magicc.create_copy()
        logger.info("Creating new magicc instance: {} - ".format(key, magicc.root_dir))
        # magicc.set_output_variables(write_binary=True, write_ascii=False)
        if magicc_version == 7:
            magicc.update_config("MAGCFG_USER.CFG", **config_defaults)
        _instances[key] = magicc
        if init_callback is not None:
            init_callback(magicc)
        else:
            magicc.set_output_variables(write_binary=True, write_ascii=False)
        return magicc


def _get_output_config(variables):
    res = {}
    var = []
    for v in variables:
        if v.startswith("CARBONCYCLE_"):
            res["out_carboncycle"] = 1
            res[
                "out_ascii_binary"
            ] = "ASCII"  # TODO: temporary until we can read carboncycle files
        elif v.startswith("DAT_"):
            var.append(v)

    # TODO: remove after pymagicc#264 is merged
    if not len(var):
        var = ["DAT_SURFACE_TEMP"]

    res["out_dynamic_vars"] = var
    return res


def read_carboncycle(fname):
    d = pd.read_csv(fname, skiprows=20, engine="python", sep=r"\s+", index_col="YEARS")
    cols = {
        "scenario": "unspecified",
        "model": "magicc",
        "region": "World",
        "unit": "unspecified",
        "variable": ["CARBONCYCLE_" + c for c in d.columns],
    }

    return MAGICCData(d, columns=cols)


def _merge_config(*ds):
    # Merge many configuration dicts case insensitively
    res = {}
    for d in ds:
        res.update({k.lower(): d[k] for k in d})
    return res


class _MAGICC_SCM(BaseSCM):
    """
    By default, all flags which control which output variables (out_*) are written to disk have been turned off. Therefore,
    explicit config variables should be set to enable the output required for training. Additionally, binary output data are
    written by default to increase the execution speed.

    One instance per process will be created. This enables the use of multiple instances of magicc to be run simultaneously.

    Parameters
    ----------
    root_dir :
        The root directory in which the magicc instances will be created. The actual instances will be
        created as subdirectories within this directory. If None, then the default directory
        in pymagicc will be used (/tmp)
    only : listlike of str or str or None
        A list of variables to extract from the magicc output. Any additional variables not in ``only`` will not be read
    kwargs :
        Additional kwargs which are passed to the pymagicc instance

    """

    def __init__(self, root_dir=None, **kwargs):
        super(_MAGICC_SCM, self).__init__(**kwargs)
        self.root_dir = root_dir

    @property
    def version(self):
        return self.magicc_cls.version

    def cleanup(self):
        vals = list(_instances.keys())
        for k in vals:
            logger.info(
                "Cleaning up magicc instance at {}".format(_instances[k].root_dir)
            )
            shutil.rmtree(_instances[k].root_dir)
            _instances.pop(k)

    def instance(self):
        return get_magicc_inst(root_dir=self.root_dir, magicc_version=self.version)

    def _run_single(self, parameters, variables=None):
        magicc = self.instance()
        if variables is None:
            variables = []
        config = _merge_config(
            self.config_parameters, parameters, _get_output_config(variables)
        )

        try:
            results = magicc.run(**config)
        except (CalledProcessError, IndexError) as e:
            logger.exception(e.stderr)
            raise ValueError(str(e)) from e

        if exists(join(magicc.out_dir, "CARBONCYCLE.OUT")):
            results = df_append(
                [results, read_carboncycle(join(magicc.out_dir, "CARBONCYCLE.OUT"))]
            )
        return results

    def _merge_config(self, *ds):
        # Merge many configuration dicts case insensitively
        res = {}
        for d in ds:
            res.update({k.lower(): d[k] for k in d})
        return res


class MAGICC6_SCM(_MAGICC_SCM):
    __doc__ = (
        """Runner for the MAGICC6 simple climate model
    """
        + _MAGICC_SCM.__doc__
    )
    magicc_cls = MAGICC6
    name = "magicc6"


class MAGICC7_SCM(_MAGICC_SCM):
    __doc__ = (
        """Runner for the MAGICC7 simple climate model
    """
        + _MAGICC_SCM.__doc__
    )
    magicc_cls = MAGICC7
    name = "magicc7"


def _parallel_func(args):
    cls, cls_kwargs, params, iter_over, variables = args
    scm = cls(**cls_kwargs)
    return scm.run(params, variables=variables, iter_over_value=iter_over)


def _init_func(instances):
    # Assigned the shared dict to _instances
    global _instances
    _instances = instances


class ParallelMagicc(_MAGICC_SCM):
    def __init__(self, n_workers=4, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._pool = None
        self.n_workers = n_workers
        self.kwargs = kwargs
        self._shared_manager = None

    def start(self):
        logger.info("Starting process pool for magicc")
        global _instances
        self._shared_manager = Manager()
        _instances = self._shared_manager.dict()

        self._pool = ProcessPoolExecutor(
            max_workers=self.n_workers, initializer=_init_func, initargs=(_instances,)
        )

    def cleanup(self):
        logger.info("Cleaning up parallel magicc instance")

        global _instances
        _instances = dict(_instances)
        if self._shared_manager is not None:
            self._shared_manager.shutdown()

        super().cleanup()

        if self._pool is not None:
            logger.info("stopping pool")
            self._pool.shutdown()

    def run_multiple(self, runs, variables=None):
        if self._pool is None:
            raise ValueError("run_multiple called outside of a context manager")
        cls = MAGICC6_SCM if self.version == 6 else MAGICC7_SCM
        runs_mp = [(cls, self.kwargs, r, i, variables) for r, i in runs]
        results = parallel_process(
            _parallel_func, runs_mp, front_num=0, pool=self._pool
        )
        for r in results:
            if not isinstance(r, ScmDataFrame):
                raise ValueError("Run failed: {}".format(r))

        return df_append(results)


class MAGICC6_Parallel(ParallelMagicc):
    """
    Parallel multi-process version of ``MAGICC6_SCM``

    Each instance is run on a worker process which allows for greater parallelisation.
    """

    magicc_cls = MAGICC6
    name = "magicc6-parallel"


class MAGICC7_Parallel(ParallelMagicc):
    """
    Parallel multi-process version of ``MAGICC7_SCM``

    Each instance is run on a worker process which allows for greater parallelisation.
    """

    magicc_cls = MAGICC7
    name = "magicc7-parallel"



import numpy as np
import pymc3.distributions as pm_dists


class Distribution(object):
    """Statistical distribution

    Used as a base for describing the distributions used to describe the possible values of the input variables.

    These Distribution's can also be converted to pymc3 Distributions if the pymc3
    optimiser is used.

    Parameters
    ----------

    Returns
    -------

    """

    def __init__(self, *args, pymc3_distribution=None, **kwargs):
        self.pymc3_distribution = pymc3_distribution
        self.args = args
        self.kwargs = kwargs

        if self.pymc3_distribution is not None:
            self.distribution = self.to_pymc3()

    def __str__(self):
        r = self.distribution._repr_latex_()
        return r if r is not None else ""

    def __getattr__(self, item):
        if item in self.kwargs:
            return self.kwargs[item]

    def evaluate(self, n):
        return self.distribution.random(size=n)

    def to_pymc3(self):
        """Allows for the custom translation to pymc3

        Parameters
        ----------

        Returns
        -------
        pymc3.Distribution
            A pymc3 distrbution
        """
        if self.pymc3_distribution is None:
            raise NotImplementedError("Cannot convert to pymc3")
        return self.pymc3_distribution.dist(*self.args, **self.kwargs)

    @classmethod
    def from_pymc3(cls, dist):
        """Distribution's created with this method will take the same parameters
        as their pymc3 equivalents

        Parameters
        ----------
        dist :
            A pymc3 distribution

        Returns
        -------
        type
            Factory for creating new instances of Distribution which wrap
            a pymc3 distribution

        """

        def create_dist(*args, **kwargs):
            return cls(*args, pymc3_distribution=dist, **kwargs)

        return create_dist


Scalar = Distribution.from_pymc3(pm_dists.Constant)
Normal = Distribution.from_pymc3(pm_dists.Normal)
LogNormal = Distribution.from_pymc3(pm_dists.Lognormal)
Uniform = Distribution.from_pymc3(pm_dists.Uniform)


class Bound(Distribution):
    def __init__(self, distribution, *args, **kwargs):
        # Extract the bounded distribution as it is handled slightly different
        self._bounded_dist = distribution
        super(Bound, self).__init__(*args, **kwargs)

        self.distribution = self.to_pymc3()

    def to_pymc3(self):
        if len(self._bounded_dist.args):
            raise ValueError(
                "Use kwargs for specifying parameters for bounded distributions"
            )
        bound = pm_dists.Bound(
            self._bounded_dist.distribution.__class__, *self.args, **self.kwargs
        )
        return bound.dist(**self._bounded_dist.kwargs)


class Function(Distribution):
    def evaluate(self, n):
        f = self.kwargs["function"]
        return np.array([f() for i in range(n)])



import scipy.optimize


from .base import BaseOptimiser
from ..distributions import Bound


class OptimiserScipy(BaseOptimiser):
    name = "scipy"

    def find_best_fit(self, evaluator, parameter_set, include_bounds=True, **kwargs):
        if "x0" in kwargs:
            x0 = kwargs.pop("x0")
        else:
            x0 = parameter_set.evaluate(1).iloc[0, :].to_dict()

        if include_bounds:
            bounds = []
            for name, p in parameter_set.tune_parameters.items():
                if isinstance(p, Bound):
                    bounds.append((p.kwargs.get("lower"), p.kwargs.get("upper")))
                else:
                    bounds.append((None, None))
            kwargs["bounds"] = bounds

        def scipy_evaluator(*args, **kwargs):
            # We want to maximise all metrics for consistency,
            # but scipy only has minimise
            return -evaluator(*args, **kwargs)

        minima = self.minimise(scipy_evaluator, parameter_set, x0, **kwargs)

        return self.process_minimisation_results(minima, parameter_set)

    def minimise(self, evaluator, parameter_set, x0, **kwargs):
        def run_function(x):
            param_dict = {n: v for n, v in zip(parameter_set.names, x)}

            return evaluator(**param_dict)

        return scipy.optimize.minimize(
            run_function, [x0[k] for k in parameter_set.names], **kwargs
        )

    def process_minimisation_results(self, res, parameter_set):
        coeffs = {k: res.x[i] for i, k in enumerate(parameter_set.names)}
        res.x = coeffs

        return res



import numpy as np
import pandas as pd
import pkg_resources
from scmdata import ScmDataFrame
from pymagicc import MAGICCData

from .base import BaseSCM


def _get_magicc6_fname(sce):
    return pkg_resources.resource_filename("pymagicc", "MAGICC6/run/{}".format(sce))


def load_emissions():
    # hack hack hack
    # only industrial emissions, lazy but will get the picture
    mdata = MAGICCData(_get_magicc6_fname("RCP85.SCEN"))
    emissions = (
        mdata.filter(
            region="World", variable="Emissions|CO2|MAGICC Fossil and Industrial"
        )
        .timeseries()
        .T.squeeze()
    )
    emissions.name = None
    emissions.index = [d.year for d in emissions.index]
    newindex = np.arange(2007, 2100)

    emissions_reindexed = emissions.reindex(index=newindex)
    emissions_reindexed.interpolate(method="linear", inplace=True)
    emissions_reindexed = pd.DataFrame(emissions_reindexed)

    mdata = MAGICCData(_get_magicc6_fname("HISTRCP_CO2I_EMIS.IN"))
    historical_emissions = mdata.timeseries().sum(axis=0)
    historical_emissions.index = mdata["year"]

    emis = pd.concat(
        [emissions_reindexed, historical_emissions]
    )  # lazy but fine for mucking around
    emis = emis.sort_index()

    return emis.index, emis.values


time_index, co2_emissions = load_emissions()


class AR5IR(BaseSCM):
    name = "ar5ir"

    a1 = 0.2240
    tau1 = 394.4
    a2 = 0.2824
    tau2 = 36.54
    a3 = 0.2763
    tau3 = 4.304
    co2_pi = 278
    f2x = 3.74
    c1 = 0.631
    d1 = 8.4
    c2 = 0.429
    d2 = 409.5
    co2_2x = False

    def __init__(self, **kwargs):
        super(AR5IR, self).__init__(**kwargs)
        for kwarg, value in kwargs.items():
            setattr(self, kwarg, value)

        self.co2_emissions = co2_emissions.copy()

        if self.co2_2x:
            self.co2_emissions = 2 * self.co2_emissions

    def _run_single(self, run_params, variables=None):
        if variables is not None and variables != ["Surface Temperature"]:
            raise ValueError("ar5ir only calculates Surface Temperature")
        # Ignoring the requested Varaibles
        for kwarg, value in run_params.items():
            setattr(self, kwarg, value)

        self.calculate_co2_concentrations()
        self.calculate_co2_radiative_forcing()
        self.calculate_temperatures()

        return ScmDataFrame(
            pd.Series(self.temperatures, index=time_index),
            columns={
                "unit": ["K"],
                "variable": ["Surface Temperature"],
                "region": ["World"],
                "model": ["ar5ir"],
                "scenario": ["unspecified"],
            },
        )

    def calculate_co2_concentrations(self):
        a0 = 1 - self.a1 - self.a2 - self.a3
        a = np.array([a0, self.a1, self.a2, self.a3])
        for a_val in a:
            assert 1 >= a_val >= 0, "that's not going to work"
        tau = np.array([10 ** 10, self.tau1, self.tau2, self.tau3])
        co2_concentration_perturbations = np.nan * np.zeros(
            (len(self.co2_emissions), 4)
        )
        co2_concentration_perturbations[0, :] = 0

        gtc_ppm_conv_factor = 18 / (
            5.14 * 12
        )  # molw.air / (m_atmos * molw.C), no idea why...
        temp = self.co2_emissions[:, np.newaxis] * a * gtc_ppm_conv_factor
        factor = 1 - 1 / tau
        for i in range(len(self.co2_emissions[:-1])):
            co2_concentration_perturbations[i + 1, :] = (  # assuming one year timestep
                co2_concentration_perturbations[i, :] * factor + temp[i]
            )

        self.co2_concentrations = (
            co2_concentration_perturbations.sum(axis=1) + self.co2_pi
        )

    def calculate_co2_radiative_forcing(self):
        self.co2_radiative_forcing = (
            self.f2x / np.log(2) * np.log(self.co2_concentrations / self.co2_pi)
        )

    def calculate_temperatures(self):
        c = np.array([self.c1, self.c2])
        d = np.array([self.d1, self.d2])
        temp_perturbations = np.nan * np.zeros((len(self.co2_emissions), 2))
        temp_perturbations[0, :] = 0

        factor = 1 - 1 / d
        temp = self.co2_radiative_forcing[:, np.newaxis] * c / d
        for i in range(len(self.co2_radiative_forcing[:-1])):
            temp_perturbations[i + 1, :] = (
                temp_perturbations[i] * factor + temp[i]
            )  # assuming one year timestep

        self.temperatures = temp_perturbations.sum(axis=1)



"""load data in the Matlab (TM) MAT-file format

Copyright (c) 2011-2018 Nephics AB
The MIT License (MIT)
"""

__all__ = ["loadmat"]


import struct
import sys
import zlib

from collections import Sequence
from itertools import tee

try:
    from itertools import izip

    ispy2 = True
except ImportError:
    izip = zip
    basestring = str
    ispy2 = False
from io import BytesIO


# encode a string to bytes and vice versa
asbytes = lambda s: s.encode("latin1")
asstr = lambda b: b.decode("latin1")

# array element data types
etypes = {
    "miINT8": {"n": 1, "fmt": "b"},
    "miUINT8": {"n": 2, "fmt": "B"},
    "miINT16": {"n": 3, "fmt": "h"},
    "miUINT16": {"n": 4, "fmt": "H"},
    "miINT32": {"n": 5, "fmt": "i"},
    "miUINT32": {"n": 6, "fmt": "I"},
    "miSINGLE": {"n": 7, "fmt": "f"},
    "miDOUBLE": {"n": 9, "fmt": "d"},
    "miINT64": {"n": 12, "fmt": "q"},
    "miUINT64": {"n": 13, "fmt": "Q"},
    "miMATRIX": {"n": 14},
    "miCOMPRESSED": {"n": 15},
    "miUTF8": {"n": 16, "fmt": "s"},
    "miUTF16": {"n": 17, "fmt": "s"},
    "miUTF32": {"n": 18, "fmt": "s"},
}

# inverse mapping of etypes
inv_etypes = dict((v["n"], k) for k, v in etypes.items())

# matrix array classes
mclasses = {
    "mxCELL_CLASS": 1,
    "mxSTRUCT_CLASS": 2,
    "mxOBJECT_CLASS": 3,
    "mxCHAR_CLASS": 4,
    "mxSPARSE_CLASS": 5,
    "mxDOUBLE_CLASS": 6,
    "mxSINGLE_CLASS": 7,
    "mxINT8_CLASS": 8,
    "mxUINT8_CLASS": 9,
    "mxINT16_CLASS": 10,
    "mxUINT16_CLASS": 11,
    "mxINT32_CLASS": 12,
    "mxUINT32_CLASS": 13,
    "mxINT64_CLASS": 14,
    "mxUINT64_CLASS": 15,
    "mxFUNCTION_CLASS": 16,
    "mxOPAQUE_CLASS": 17,
    "mxOBJECT_CLASS_FROM_MATRIX_H": 18,
}

# map of numeric array classes to data types
numeric_class_etypes = {
    "mxDOUBLE_CLASS": "miDOUBLE",
    "mxSINGLE_CLASS": "miSINGLE",
    "mxINT8_CLASS": "miINT8",
    "mxUINT8_CLASS": "miUINT8",
    "mxINT16_CLASS": "miINT16",
    "mxUINT16_CLASS": "miUINT16",
    "mxINT32_CLASS": "miINT32",
    "mxUINT32_CLASS": "miUINT32",
    "mxINT64_CLASS": "miINT64",
    "mxUINT64_CLASS": "miUINT64",
}

inv_mclasses = dict((v, k) for k, v in mclasses.items())

# data types that may be used when writing numeric data
compressed_numeric = ["miINT32", "miUINT16", "miINT16", "miUINT8"]


def diff(iterable):
    """Diff elements of a sequence:
    s -> s0 - s1, s1 - s2, s2 - s3, ...
    """
    a, b = tee(iterable)
    next(b, None)
    return (i - j for i, j in izip(a, b))


#
# Uitlity functions
#


def unpack(endian, fmt, data):
    """Unpack a byte string to the given format. If the byte string
    contains more bytes than required for the given format, the function
    returns a tuple of values.
    """
    if fmt == "s":
        # read data as an array of chars
        val = struct.unpack("".join([endian, str(len(data)), "s"]), data)[0]
    else:
        # read a number of values
        num = len(data) // struct.calcsize(fmt)
        val = struct.unpack("".join([endian, str(num), fmt]), data)
        if len(val) == 1:
            val = val[0]
    return val


def read_file_header(fd, endian):
    """Read mat 5 file header of the file fd.
    Returns a dict with header values.
    """
    fields = [
        ("description", "s", 116),
        ("subsystem_offset", "s", 8),
        ("version", "H", 2),
        ("endian_test", "s", 2),
    ]
    hdict = {}
    for name, fmt, num_bytes in fields:
        data = fd.read(num_bytes)
        hdict[name] = unpack(endian, fmt, data)
    hdict["description"] = hdict["description"].strip()
    v_major = hdict["version"] >> 8
    v_minor = hdict["version"] & 0xFF
    hdict["__version__"] = "%d.%d" % (v_major, v_minor)
    return hdict


def read_element_tag(fd, endian):
    """Read data element tag: type and number of bytes.
    If tag is of the Small Data Element (SDE) type the element data
    is also returned.
    """
    data = fd.read(8)
    mtpn = unpack(endian, "I", data[:4])
    # The most significant two bytes of mtpn will always be 0,
    # if they are not, this must be SDE format
    num_bytes = mtpn >> 16
    if num_bytes > 0:
        # small data element format
        mtpn = mtpn & 0xFFFF
        if num_bytes > 4:
            raise ParseError("Error parsing Small Data Element (SDE) " "formatted data")
        data = data[4 : 4 + num_bytes]
    else:
        # regular element
        num_bytes = unpack(endian, "I", data[4:])
        data = None
    return (mtpn, num_bytes, data)


def read_elements(fd, endian, mtps, is_name=False):
    """Read elements from the file.

    If list of possible matrix data types mtps is provided, the data type
    of the elements are verified.
    """
    mtpn, num_bytes, data = read_element_tag(fd, endian)
    if mtps and mtpn not in [etypes[mtp]["n"] for mtp in mtps]:
        raise ParseError(
            "Got type {}, expected {}".format(
                mtpn,
                " / ".join("{} ({})".format(etypes[mtp]["n"], mtp) for mtp in mtps),
            )
        )
    if not data:
        # full format, read data
        data = fd.read(num_bytes)
        # Seek to next 64-bit boundary
        mod8 = num_bytes % 8
        if mod8:
            fd.seek(8 - mod8, 1)

    # parse data and return values
    if is_name:
        # names are stored as miINT8 bytes
        fmt = "s"
        val = [unpack(endian, fmt, s) for s in data.split(b"\0") if s]
        if len(val) == 0:
            val = ""
        elif len(val) == 1:
            val = asstr(val[0])
        else:
            val = [asstr(s) for s in val]
    else:
        fmt = etypes[inv_etypes[mtpn]]["fmt"]
        val = unpack(endian, fmt, data)
    return val


def read_header(fd, endian):
    """Read and return the matrix header."""
    flag_class, nzmax = read_elements(fd, endian, ["miUINT32"])
    header = {
        "mclass": flag_class & 0x0FF,
        "is_logical": (flag_class >> 9 & 1) == 1,
        "is_global": (flag_class >> 10 & 1) == 1,
        "is_complex": (flag_class >> 11 & 1) == 1,
        "nzmax": nzmax,
    }
    header["dims"] = read_elements(fd, endian, ["miINT32"])
    header["n_dims"] = len(header["dims"])
    if header["n_dims"] != 2:
        raise ParseError("Only matrices with dimension 2 are supported.")
    header["name"] = read_elements(fd, endian, ["miINT8"], is_name=True)
    return header


def read_var_header(fd, endian):
    """Read full header tag.

    Return a dict with the parsed header, the file position of next tag,
    a file like object for reading the uncompressed element data.
    """
    mtpn, num_bytes = unpack(endian, "II", fd.read(8))
    next_pos = fd.tell() + num_bytes

    if mtpn == etypes["miCOMPRESSED"]["n"]:
        # read compressed data
        data = fd.read(num_bytes)
        dcor = zlib.decompressobj()
        # from here, read of the decompressed data
        fd_var = BytesIO(dcor.decompress(data))
        del data
        fd = fd_var
        # Check the stream is not so broken as to leave cruft behind
        if dcor.flush() != b"":
            raise ParseError("Error in compressed data.")
        # read full tag from the uncompressed data
        mtpn, num_bytes = unpack(endian, "II", fd.read(8))

    if mtpn != etypes["miMATRIX"]["n"]:
        raise ParseError(
            "Expecting miMATRIX type number {}, "
            "got {}".format(etypes["miMATRIX"]["n"], mtpn)
        )
    # read the header
    header = read_header(fd, endian)
    return header, next_pos, fd


def squeeze(array):
    """Return array contents if array contains only one element.
    Otherwise, return the full array.
    """
    if len(array) == 1:
        array = array[0]
    return array


def read_numeric_array(fd, endian, header, data_etypes):
    """Read a numeric matrix.
    Returns an array with rows of the numeric matrix.
    """
    if header["is_complex"]:
        raise ParseError("Complex arrays are not supported")
    # read array data (stored as column-major)
    data = read_elements(fd, endian, data_etypes)
    if not isinstance(data, Sequence):
        # not an array, just a value
        return data
    # transform column major data continous array to
    # a row major array of nested lists
    rowcount = header["dims"][0]
    colcount = header["dims"][1]
    array = [
        list(data[c * rowcount + r] for c in range(colcount)) for r in range(rowcount)
    ]
    # pack and return the array
    return squeeze(array)


def read_cell_array(fd, endian, header):
    """Read a cell array.
    Returns an array with rows of the cell array.
    """
    array = [list() for i in range(header["dims"][0])]
    for row in range(header["dims"][0]):
        for col in range(header["dims"][1]):
            # read the matrix header and array
            vheader, next_pos, fd_var = read_var_header(fd, endian)
            varray = read_var_array(fd_var, endian, vheader)
            array[row].append(varray)
            # move on to next field
            fd.seek(next_pos)
    # # pack and return the array
    # if header['dims'][0] == 1:
    #     return squeeze(array[0])
    return squeeze(array)


def read_struct_array(fd, endian, header):
    """Read a struct array.
    Returns a dict with fields of the struct array.
    """
    # read field name length (unused, as strings are null terminated)
    field_name_length = read_elements(fd, endian, ["miINT32"])
    if field_name_length > 32:
        raise ParseError("Unexpected field name length: {}".format(field_name_length))

    # read field names
    fields = read_elements(fd, endian, ["miINT8"], is_name=True)
    if isinstance(fields, basestring):
        fields = [fields]

    # read rows and columns of each field
    empty = lambda: [list() for i in range(header["dims"][0])]
    array = {}
    for row in range(header["dims"][0]):
        for col in range(header["dims"][1]):
            for field in fields:
                # read the matrix header and array
                vheader, next_pos, fd_var = read_var_header(fd, endian)
                data = read_var_array(fd_var, endian, vheader)
                if field not in array:
                    array[field] = empty()
                array[field][row].append(data)
                # move on to next field
                fd.seek(next_pos)
    # pack the nested arrays
    for field in fields:
        rows = array[field]
        for i in range(header["dims"][0]):
            rows[i] = squeeze(rows[i])
        array[field] = squeeze(array[field])
    return array


def read_char_array(fd, endian, header):
    array = read_numeric_array(fd, endian, header, ["miUTF8"])
    if header["dims"][0] > 1:
        # collapse rows of chars into a list of strings
        array = [asstr(bytearray(i)) for i in array]
    else:
        # collaps row of chars into a single string
        array = asstr(bytearray(array))
    return array


def read_var_array(fd, endian, header):
    """Read variable array (of any supported type)."""
    mc = inv_mclasses[header["mclass"]]

    if mc in numeric_class_etypes:
        return read_numeric_array(
            fd,
            endian,
            header,
            set(compressed_numeric).union([numeric_class_etypes[mc]]),
        )
    elif mc == "mxSPARSE_CLASS":
        raise ParseError("Sparse matrices not supported")
    elif mc == "mxCHAR_CLASS":
        return read_char_array(fd, endian, header)
    elif mc == "mxCELL_CLASS":
        return read_cell_array(fd, endian, header)
    elif mc == "mxSTRUCT_CLASS":
        return read_struct_array(fd, endian, header)
    elif mc == "mxOBJECT_CLASS":
        raise ParseError("Object classes not supported")
    elif mc == "mxFUNCTION_CLASS":
        raise ParseError("Function classes not supported")
    elif mc == "mxOPAQUE_CLASS":
        raise ParseError("Anonymous function classes not supported")


def eof(fd):
    """Determine if end-of-file is reached for file fd."""
    b = fd.read(1)
    end = len(b) == 0
    if not end:
        curpos = fd.tell()
        fd.seek(curpos - 1)
    return end


class ParseError(Exception):
    pass


#
# Read from MAT file
#


def loadmat(filename, meta=False):
    """Load data from MAT-file:

    data = loadmat(filename, meta=False)

    The filename argument is either a string with the filename, or
    a file like object.

    The returned parameter ``data`` is a dict with the variables found
    in the MAT file.

    Call ``loadmat`` with parameter meta=True to include meta data, such
    as file header information and list of globals.

    A ``ParseError`` exception is raised if the MAT-file is corrupt or
    contains a data type that cannot be parsed.
    """

    if isinstance(filename, basestring):
        fd = open(filename, "rb")
    else:
        fd = filename

    # Check mat file format is version 5
    # For 5 format we need to read an integer in the header.
    # Bytes 124 through 128 contain a version integer and an
    # endian test string
    fd.seek(124)
    tst_str = fd.read(4)
    little_endian = tst_str[2:4] == b"IM"
    endian = ""
    if (sys.byteorder == "little" and little_endian) or (
        sys.byteorder == "big" and not little_endian
    ):
        # no byte swapping same endian
        pass
    elif sys.byteorder == "little":
        # byte swapping
        endian = ">"
    else:
        # byte swapping
        endian = "<"
    maj_ind = int(little_endian)
    # major version number
    maj_val = ord(tst_str[maj_ind]) if ispy2 else tst_str[maj_ind]
    if maj_val != 1:
        raise ParseError("Can only read from Matlab level 5 MAT-files")
    # the minor version number (unused value)
    # min_val = ord(tst_str[1 - maj_ind]) if ispy2 else tst_str[1 - maj_ind]

    mdict = {}
    if meta:
        # read the file header
        fd.seek(0)
        mdict["__header__"] = read_file_header(fd, endian)
        mdict["__globals__"] = []

    # read data elements
    while not eof(fd):
        hdr, next_position, fd_var = read_var_header(fd, endian)
        name = hdr["name"]
        if name in mdict:
            raise ParseError('Duplicate variable name "{}" in mat file.'.format(name))

        # read the matrix
        mdict[name] = read_var_array(fd_var, endian, hdr)
        if meta and hdr["is_global"]:
            mdict["__globals__"].append(name)

        # move on to next entry in file
        fd.seek(next_position)

    fd.close()
    return mdict



from concurrent.futures import ProcessPoolExecutor, as_completed, wait
from copy import deepcopy
from datetime import datetime
from logging import getLogger

import numpy as np
import pandas as pd
from scmdata import ScmDataFrame
from tqdm import tqdm

from . import mat4py

logger = getLogger(__name__)


# ZN: can we get rid of most of these given the new capabilities of ScmDataFrame?
def load_iam_timeseries(fname):
    # Read in the data. It requires a bit of wrangling to get it in the IamDataFrame format
    df = pd.read_csv(fname, header=None)
    df = df.set_index(0).T
    df.pop("time")
    df = ScmDataFrame(df)
    df["time"] = [
        datetime(d.year, d.month, d.day) for d in df["time"]
    ]  # strip out any hour values
    df = df.filter(
        region=[
            "World",
            "World|Northern Hemisphere|Land",
            "World|Northern Hemisphere|Ocean",
            "World|Southern Hemisphere|Land",
            "World|Southern Hemisphere|Ocean",
        ]
    )

    # Resample to annual means
    return df.resample("AS")


def merge_iam_timeseries(*args):
    """Merges two or more timeseries in the `ScmDataFrame` format.

    This method assumes that all data are from the same scenario. All scenarios in the resulting dataframe will be the same as specified
    by data in the last value of ``dfs``

    Parameters
    ----------
    args : :obj:`scmdata.ScmDataFrame`or string
        2 or more items to merge. :obj:`scmdata.ScmDataFrame` objects can be passed directly otherwise, if a string is passed
        then the dataframe will be attempted to be read from file

    Returns
    -------
    :obj:`scmdata.ScmDataFrame`

    """
    dfs = args
    assert len(dfs) >= 2, "At least two input dfs are required"

    def get_df(f):
        if isinstance(f, ScmDataFrame):
            return f
        else:
            return load_iam_timeseries(f)

    # Ensure that each df is loaded
    dfs = [get_df(f) for f in dfs]
    target_scenario = dfs[-1]["scenario"].unique()[0]
    res = None
    for df in dfs:
        df_copy = deepcopy(df)

        df_copy["scenario"] = target_scenario
        if res is None:
            res = df_copy
        else:
            res = res.append(df_copy)

    return res


def create_iam_dataframe(
    ts,
    index=None,
    variable="Surface Temperature",
    scenario="unspecified",
    model="unspecified",
    region="World",
    unit="K",
):
    """Create a :obj:`scmdata.ScmDataFrame` from a timeseries

    A number of the interfaces within scmcallib expect :obj:`scmdata.ScmDataFrame`, but they are not always the easiest things to initialise. This
    function provides a helper function for converting a 1d timeseries into a :obj:`scmdata.ScmDataFrame`

    Parameters
    ----------
    ts :
        1D np array or pd.Series/pd.Dataframe.
    index :
        If None, use the index from the ts. This assumes that `ts` is a pd.Series or pd.DataFrame. (Default value = None)
    variable :
         (Default value = 'Surface Temperature')
    scenario :
         (Default value = 'unspecified')
    model :
         (Default value = 'unspecified')
    region :
         (Default value = 'World')
    unit :
         (Default value = 'K')

    Returns
    -------
    :obj:`scmdata.ScmDataFrame`
    """
    if index is None:
        if isinstance(ts, (pd.DataFrame, pd.Series)):
            index = ts.index
        else:
            raise ValueError("No valid index")

    if isinstance(ts, (pd.DataFrame, pd.Series)):
        assert len(ts)
        ts = ts.values.squeeze()

    assert ts.ndim == 1
    return ScmDataFrame(
        pd.Series(ts, index=index),
        columns={
            "unit": [unit],
            "variable": [variable],
            "region": [region],
            "model": [model],
            "scenario": [scenario],
        },
    )


def prepare_dataframe(df, extra_cols=()):
    """Convert model output in the form of a ScmDataFrame into something more useful in this use case

    Parameters
    ----------
    df : :obj:`scmdata.ScmDataFrame`
        Timeseries data from an scm or other training data
    extra_cols : listlike of str
        Additional meta columns to include in the output DataFrame

    Returns
    -------
    pd.DataFrame
        DataFrame with a time index and MultiIndex columns for model and region

    """
    return df.timeseries(meta=["variable", "region"] + list(extra_cols)).T


def convert_tuningstruc_to_scmdf(
    filepath, variable=None, region=None, unit=None, scenario=None, model=None
):
    """Convert a matlab tuningstruc to an ScmDataFrame

    Parameters
    ----------
    filepath : str
        Filepath from which to load the data

    variable : str
        Name of the variable contained in the tuningstruc. If None,
        `convert_tuningstruc_to_scmdf` will attempt to determine it from the input
        file.

    region : str
        Region to which the data in the tuningstruc applies. If None,
        `convert_tuningstruc_to_scmdf` will attempt to determine it from the input
        file.

    unit : str
        Units of the data in the tuningstruc. If None,
        `convert_tuningstruc_to_scmdf` will attempt to determine it from the input
        file.

    scenario : str
        Scenario to which the data in the tuningstruc applies. If None,
        `convert_tuningstruc_to_scmdf` will attempt to determine it from the input
        file.

    model : str
        The (integrated assessment) model which generated the emissions scenario
        associated with the data in the tuningstruc. If None,
        `convert_tuningstruc_to_scmdf` will attempt to determine it from the input
        file and if it cannot, it will be set to "unspecified".

    Raises
    ------
    ValueError
        If a metadata variable is not supplied and it cannot be determined from the
        tuningstruc.

    Returns
    -------
    :obj: `ScmDataFrame`
        ScmDataFrame with the tuningstruc data
    """
    dataset = mat4py.loadmat(filepath)

    for m, climate_model in enumerate(dataset["tuningdata"]["modelcodes"]):
        metadata = {
            "variable": [variable],
            "region": [region],
            "unit": [unit],
            "climate_model": [climate_model],
            "scenario": [scenario],
            "model": [model],
        }
        for k, v in metadata.items():
            if v == [None]:
                try:
                    metadata[k] = [dataset["tuningdata"]["model"][m][k]]
                except KeyError:
                    if k == "model":
                        metadata[k] = ["unspecified"]
                        continue

                    error_msg = "Cannot determine {} from file: " "{}".format(
                        k, filepath
                    )
                    raise KeyError(error_msg)

        scmdf = ScmDataFrame(
            data=np.asarray(dataset["tuningdata"]["model"][m]["data"][1]),
            columns=metadata,
            index=dataset["tuningdata"]["model"][m]["data"][0],
        )
        try:
            ref_df.append(scmdf, inplace=True)
        except NameError:
            ref_df = scmdf

    return ref_df


def convert_scmdf_to_tuningstruc(scmdf, outpath):
    """Convert an ScmDataFrame to a matlab tuningstruc

    One tuningstruc file will be created for each unique
    ["model", "scenario", "variable", "region", "unit"] combination in the input
    ScmDataFrame.

    Parameters
    ----------
    scmdf : :obj: `ScmDataFrame`
        ScmDataFrame to convert to a tuningstruc

    outpath : str
        Base path in which to save the tuningstruc. The rest of the pathname is
        generated from the metadata. `.mat` is also appended automatically.
    """
    iter = scmdf.timeseries().groupby(
        ["model", "scenario", "variable", "region", "unit"]
    )
    for (model, scenario, variable, region, unit), df in iter:
        dataset = {}
        dataset["tuningdata"] = {}
        dataset["tuningdata"]["modelcodes"] = []
        dataset["tuningdata"]["model"] = []

        for m, (climate_model, df) in enumerate(df.groupby("climate_model")):
            # impossible to make dataframe with duplicate rows so not an issue
            # this is just in case
            error_msg = (
                "Should only have a single unique timeseries for a given "
                '["climate_model", "model", "scenario", "variable", '
                '"region", "unit"] combination'
            )
            assert df.shape[0] == 1, error_msg

            dataset["tuningdata"]["modelcodes"].append(climate_model)
            dataset["tuningdata"]["model"].append({})

            dataset["tuningdata"]["model"][m]["model"] = model
            dataset["tuningdata"]["model"][m]["scenario"] = scenario
            dataset["tuningdata"]["model"][m]["variable"] = variable
            dataset["tuningdata"]["model"][m]["region"] = region
            dataset["tuningdata"]["model"][m]["unit"] = unit

            dataset["tuningdata"]["model"][m]["notes"] = (
                "{} {} {} {} ({}) tuningstruc (written with scmcallib)"
                "".format(scenario, model, region, variable, unit)
            )
            dataset["tuningdata"]["model"][m]["data"] = [
                [float(t.year) for t in df.columns],
                list(df.values.squeeze()),
            ]
            dataset["tuningdata"]["model"][m]["col_code"] = ["YEARS", variable]

        outfile = (
            "{}_{}_{}_{}_{}.mat".format(outpath, scenario, model, variable, region)
            .replace(" ", "_")
            .replace("|", "_")
        )
        # ask Jared how to add logging here
        mat4py.savemat(outfile, dataset)


def parallel_process(
    func, arr, n_jobs=10, use_kwargs=False, front_num=3, show_progress=False, pool=None
):
    """
    A parallel version of the map function with a progress bar.

    Adapted from http://danshiebler.com/2016-09-14-parallel-progress-bar/

    Parameters
    ----------
    func : function
        A python function to apply to the elements of array
    arr: listlike
        An array of configuration to iterate over
    n_jobs: int default=16
    use_kwargs : boolean, default=False
        Whether to consider the elements of array as dictionaries of keyword arguments to function
    front_num : int, default=3
        The number of iterations to run serially before kicking off the parallel job. Useful for catching bugs

    Returns
    -------
        [func(arr[0]), func(arr[1]), ...]
    """

    # We run the first few iterations serially to catch bugs
    front = []
    if front_num > 0:
        front = [func(**a) if use_kwargs else func(a) for a in arr[:front_num]]
    # If we set n_jobs to 1, just run a list comprehension. This is useful for benchmarking and debugging.
    if n_jobs == 1:
        if show_progress:
            return front + [
                func(**a) if use_kwargs else func(a) for a in tqdm(arr[front_num:])
            ]
        else:
            return front + [
                func(**a) if use_kwargs else func(a) for a in arr[front_num:]
            ]

    cleanup_pool = False
    if pool is None:
        pool = ProcessPoolExecutor(n_jobs)
        cleanup_pool = True

    try:
        # Pass the elements of array into func
        if use_kwargs:
            futures = [pool.submit(func, **a) for a in arr[front_num:]]
        else:
            futures = [pool.submit(func, a) for a in arr[front_num:]]
        kwargs = {
            "total": len(futures),
            "unit": "it",
            "unit_scale": True,
            "leave": True,
        }
        # Print out the progress as tasks complete
        if show_progress:
            for f in tqdm(as_completed(futures), **kwargs):
                pass
        else:
            wait(futures)
    finally:
        if cleanup_pool:
            pool.shutdown()

    out = []
    # Get the results from the futures.
    for i, future in enumerate(futures):
        try:
            out.append(future.result())
        except Exception as e:
            out.append(e)
    return front + out
